from .base import _BaseLiNGAM
from .utils import check_array

from tqdm import tqdm
import numpy as np
from lingam_cuda import causal_order

class DirectLiNGAM(_BaseLiNGAM):
    """Implementation of DirectLiNGAM Algorithm [1]_ [2]_

    References
    ----------
    .. [1] S. Shimizu, T. Inazumi, Y. Sogawa, A. Hyvärinen, Y. Kawahara, T. Washio, P. O. Hoyer and K. Bollen.
       DirectLiNGAM: A direct method for learning a linear non-Gaussian structural equation model. Journal of Machine Learning Research, 12(Apr): 1225--1248, 2011.
    .. [2] A. Hyvärinen and S. M. Smith. Pairwise likelihood ratios for estimation of non-Gaussian structural eauation models.
       Journal of Machine Learning Research 14:111-152, 2013.
    """

    def __init__(self, random_state=None, prior_knowledge=None, apply_prior_knowledge_softly=False, measure='pwling'):
        """Construct a DirectLiNGAM model.

        Parameters
        ----------
        random_state : int, optional (default=None)
            ``random_state`` is the seed used by the random number generator.
        prior_knowledge : array-like, shape (n_features, n_features), optional (default=None)
            Prior background_knowledge used for causal discovery, where ``n_features`` is the number of features.

            The elements of prior background_knowledge matrix are defined as follows [1]_:

            * ``0`` : :math:`x_i` does not have a directed path to :math:`x_j`
            * ``1`` : :math:`x_i` has a directed path to :math:`x_j`
            * ``-1`` : No prior background_knowledge is available to know if either of the two cases above (0 or 1) is true.
        apply_prior_knowledge_softly : boolean, optional (default=False)
            If True, apply prior background_knowledge softly.
        measure : {'pwling', 'kernel'}, optional (default='pwling')
            Measure to evaluate independence: typically 'pwling' [2]_ or 'kernel' [1]_ but only pwling supported.
        """
        super().__init__(random_state)
        self._Aknw = prior_knowledge
        self._apply_prior_knowledge_softly = apply_prior_knowledge_softly
        self._measure = measure

        if self._Aknw is not None:
            self._Aknw = check_array(self._Aknw)
            self._Aknw = np.where(self._Aknw < 0, np.nan, self._Aknw)

            # Extract all partial orders in prior background_knowledge matrix
            if not self._apply_prior_knowledge_softly:
                self._partial_orders = self._extract_partial_orders(self._Aknw)

    def _search_causal_order(self, X, U):
        """Accelerated Causal ordering.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where ``n_samples`` is the number of samples
            and ``n_features`` is the number of features.
        U: indices of cols in X

        Returns
        -------
        self : object
            Returns the instance itself.
        mlist: causal ordering
        """
        cols = len(U)
        rows = len(X)

        arr = X[:, np.array(U)]
        mlist = causal_order(arr, rows, cols)
        return U[np.argmax(mlist)]

    def _entropy(self, u):
        """Calculate entropy using the maximum entropy approximations."""
        k1 = 79.047
        k2 = 7.4129
        gamma = 0.37457
        return (1 + np.log(2 * np.pi)) / 2 - \
               k1 * (np.mean(np.log(np.cosh(u))) - gamma) ** 2 - \
               k2 * (np.mean(u * np.exp((-u ** 2) / 2))) ** 2

    def _diff_mutual_info(self, xi_std, xj_std, ri_j, rj_i):
        """Calculate the difference of the mutual information."""

        return (self._entropy(xj_std) + self._entropy(ri_j / np.std(ri_j))) - \
               (self._entropy(xi_std) + self._entropy(rj_i / np.std(rj_i)))


    def _search_causal_order_(self, X, U):
        """Search the causal ordering. Sequential"""
        Uc, Vj = self._search_candidate(U)
        if len(Uc) == 1:
            return Uc[0]

        M_list = []
        for i in Uc:
            M = 0
            for j in U:
                if i != j:
                    xi_std = (X[:, i] - np.mean(X[:, i])) / np.std(X[:, i])
                    xj_std = (X[:, j] - np.mean(X[:, j])) / np.std(X[:, j])
                    ri_j = xi_std if i in Vj and j in Uc else self._residual(
                        xi_std, xj_std)
                    rj_i = xj_std if j in Vj and i in Uc else self._residual(
                        xj_std, xi_std)
                    M += np.min([0, self._diff_mutual_info(xi_std,
                                                           xj_std, ri_j, rj_i)]) ** 2
            M_list.append(-1.0 * M)

        return Uc[np.argmax(M_list)]


    def fit(self, X, disable_tqdm=False):
        """Fit the model to X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where ``n_samples`` is the number of samples
            and ``n_features`` is the number of features.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        # Check parameters
        X = check_array(X)
        n_features = X.shape[1]
        if self._Aknw is not None:
            if (n_features, n_features) != self._Aknw.shape:
                raise ValueError(
                    'The shape of prior background_knowledge must be (n_features, n_features)')

        # Causal discovery
        U = np.arange(n_features).astype(np.int32)
        K = []
        X_ = np.copy(X).astype(np.float64)
        for _ in tqdm(range(n_features), disable=disable_tqdm):

            m = self._search_causal_order(X_, U)
            U = U[U != m]
            # for i in U:
            #     X_[:, i] = self._residual(X_[:, i], X_[:, m])
            X_ = self.vec_residual(X_, m)

            K.append(m)

            # Update partial orders
            if (self._Aknw is not None) and (not self._apply_prior_knowledge_softly):
                self._partial_orders = self._partial_orders[self._partial_orders[:, 0] != m]

        self._causal_order = K
        return self._estimate_adjacency_matrix(X, prior_knowledge=self._Aknw)


    def _extract_partial_orders(self, pk):
        """ Extract partial orders from prior background_knowledge."""
        path_pairs = np.array(np.where(pk == 1)).transpose()
        no_path_pairs = np.array(np.where(pk == 0)).transpose()

        # Check for inconsistencies in pairs with path
        check_pairs = np.concatenate([path_pairs, path_pairs[:, [1, 0]]])
        if len(check_pairs) > 0:
            pairs, counts = np.unique(check_pairs, axis=0, return_counts=True)
            if len(pairs[counts > 1]) > 0:
                raise ValueError(
                    f'The prior background_knowledge contains inconsistencies at the following indices: {pairs[counts > 1].tolist()}')

        # Check for inconsistencies in pairs without path.
        # If there are duplicate pairs without path, they cancel out and are not ordered.
        check_pairs = np.concatenate([no_path_pairs, no_path_pairs[:, [1, 0]]])
        if len(check_pairs) > 0:
            pairs, counts = np.unique(check_pairs, axis=0, return_counts=True)
            check_pairs = np.concatenate([no_path_pairs, pairs[counts > 1]])
            pairs, counts = np.unique(check_pairs, axis=0, return_counts=True)
            no_path_pairs = pairs[counts < 2]

        check_pairs = np.concatenate([path_pairs, no_path_pairs[:, [1, 0]]])
        if len(check_pairs) == 0:
            # If no pairs are extracted from the specified prior background_knowledge,
            # discard the prior background_knowledge.
            self._Aknw = None
            return None

        pairs = np.unique(check_pairs, axis=0)
        return pairs[:, [1, 0]]  # [to, from] -> [from, to]


    def _residual(self, xi, xj):
        """The residual when xi is regressed on xj."""
        return xi - (np.cov(xi, xj)[0, 1] / np.var(xj)) * xj


    def vec_residual(self, X_, m):
        """The residual when xi is regressed on xj."""

        x_m = X_[:, m][:, None]
        cov_matrix = np.cov(X_, x_m, rowvar=False, bias=True)
        cov_with_m = cov_matrix[:-1, -1]
        var_m = cov_matrix[-1, -1]
        return  X_ - (cov_with_m / var_m) * x_m


    def _search_candidate(self, U):
        """ Search for candidate features """
        # If no prior background_knowledge is specified, nothing to do.
        if self._Aknw is None:
            return U, []

        # Apply prior background_knowledge in a strong way
        if not self._apply_prior_knowledge_softly:
            Uc = [i for i in U if i not in self._partial_orders[:, 1]]
            return Uc, []

        # Find exogenous features
        Uc = []
        for j in U:
            index = U[U != j]
            if self._Aknw[j][index].sum() == 0:
                Uc.append(j)

        # Find endogenous features, and then find candidate features
        if len(Uc) == 0:
            U_end = []
            for j in U:
                index = U[U != j]
                if np.nansum(self._Aknw[j][index]) > 0:
                    U_end.append(j)

            # Find sink features (original)
            for i in U:
                index = U[U != i]
                if self._Aknw[index, i].sum() == 0:
                    U_end.append(i)
            Uc = [i for i in U if i not in set(U_end)]

        # make V^(j)
        Vj = []
        for i in U:
            if i in Uc:
                continue
            if self._Aknw[i][Uc].sum() == 0:
                Vj.append(i)
        return Uc, Vj



from .utils import check_array, resample, hsic_test_gamma, predict_adaptive_lasso, scale, BootstrapResult
from .base import _BaseLiNGAM
from .directlingam import DirectLiNGAM

import numpy as np
import warnings
import itertools
from statsmodels.tsa.vector_ar.var_model import VAR

class VARLiNGAM:
    """Implementation of VAR-LiNGAM Algorithm [1]_

    References
    ----------
    .. [1] Aapo Hyvärinen, Kun Zhang, Shohei Shimizu, Patrik O. Hoyer.
       Estimation of a Structural Vector Autoregression Model Using Non-Gaussianity. Journal of Machine Learning Research, 11: 1709-1731, 2010.
    """

    def __init__(self, lags=1, criterion='bic', prune=True, ar_coefs=None, lingam_model=None, random_state=None):
        """Construct a VARLiNGAM model.

        Parameters
        ----------
        lags : int, optional (default=1)
            Number of lags.
        criterion : {‘aic’, ‘fpe’, ‘hqic’, ‘bic’, None}, optional (default='bic')
            Criterion to decide the best lags within ``lags``.
            Searching the best lags is disabled if ``criterion`` is ``None``.
        prune : boolean, optional (default=True)
            Whether to prune the adjacency matrix of lags or not.
        ar_coefs : array-like, optional (default=None)
            Coefficients of AR model. Estimating AR model is skipped if specified ``ar_coefs``.
            Shape must be (``lags``, n_features, n_features).
        lingam_model : lingam object inherits 'lingam._BaseLiNGAM', optional (default=None)
            LiNGAM model for causal discovery. If None, DirectLiNGAM algorithm is selected.
        random_state : int, optional (default=None)
            ``random_state`` is the seed used by the random number generator.
        """
        self._lags = lags
        self._criterion = criterion
        self._prune = prune
        self._ar_coefs = check_array(
            ar_coefs, allow_nd=True) if ar_coefs is not None else None
        self._lingam_model = lingam_model
        self._random_state = random_state

    def fit(self, X, disable_tqdm=False, original=False):
        """Fit the model to X.

        Parameters
        ----------
        X: array-like, shape (n_samples, n_features)
            Training data, where ``n_samples`` is the number of samples
            and ``n_features`` is the number of features.

        returns
        -------
        self : object
            Returns the instance itself.
        """
        self._causal_order = None
        self._adjacency_matrices = None

        X = check_array(X)
        # X = scale(X)

        lingam_model = self._lingam_model
        if lingam_model is None:
            lingam_model = DirectLiNGAM()
        elif not isinstance(lingam_model, _BaseLiNGAM):
            raise ValueError('lingam_model must be a subclass of _BaseLiNGAM')

        M_taus = self._ar_coefs

        if M_taus is None:
            M_taus, lags, residuals = self._estimate_var_coefs(X)
        else:
            lags = M_taus.shape[0]
            residuals = self._calc_residuals(X, M_taus, lags)

        model = lingam_model
        model.fit(residuals, disable_tqdm=disable_tqdm, original=original)

        B_taus = self._calc_b(X, model._adjacency_matrix, M_taus)

        if self._prune:
            B_taus = self._pruning(X, B_taus, model._causal_order)

        self._ar_coefs = M_taus
        self._lags = lags
        self._residuals = residuals

        self._causal_order = model._causal_order
        self._adjacency_matrix = B_taus

        return self

    def bootstrap(self, X, n_sampling):
        """Evaluate the statistical reliability of DAG based on the bootstrapping.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where ``n_samples`` is the number of samples
            and ``n_features`` is the number of features.
        n_sampling : int
            Number of bootstrapping samples.

        Returns
        -------
        result : TimeseriesBootstrapResult
            Returns the result of bootstrapping.
        """
        X = check_array(X)

        n_samples = X.shape[0]
        n_features = X.shape[1]
        lags = self._lags

        criterion = self._criterion
        self._criterion = None

        self.fit(X)

        residuals = self._residuals
        ar_coefs = self._ar_coefs

        total_effects = np.zeros(
            [n_sampling, n_features, n_features * (1 + self._lags)])

        adjacency_matrices = []
        for i in range(n_sampling):
            sampled_residuals = resample(residuals, n_samples=n_samples)

            resampled_X = np.zeros((n_samples, n_features))
            for j in range(n_samples):
                if j < lags:
                    resampled_X[j, :] = sampled_residuals[j]
                    continue

                ar = np.zeros((1, n_features))
                for t, M in enumerate(ar_coefs):
                    ar += np.dot(M, resampled_X[j - t - 1, :].T).T

                resampled_X[j, :] = ar + sampled_residuals[j]

            self.fit(resampled_X)
            am = np.concatenate([*self._adjacency_matrices], axis=1)
            adjacency_matrices.append(am)

            # total effects
            for c, to in enumerate(reversed(self._causal_order)):
                # time t
                for from_ in self._causal_order[:n_features - (c + 1)]:
                    total_effects[i, to, from_] = self.estimate_total_effect(
                        resampled_X, from_, to)

                # time t-tau
                for lag in range(self._lags):
                    for from_ in range(n_features):
                        total_effects[i, to, from_ + n_features] = self.estimate_total_effect(
                            resampled_X, from_, to, lag + 1)

        self._criterion = criterion

        return BootstrapResult(adjacency_matrices, total_effects)

    def estimate_total_effect(self, X, from_index, to_index, from_lag=0):
        """Estimate total effect using causal model.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Original data, where n_samples is the number of samples
            and n_features is the number of features.
        from_index :
            Index of source variable to estimate total effect.
        to_index :
            Index of destination variable to estimate total effect.

        Returns
        -------
        total_effect : float
            Estimated total effect.
        """
        X = check_array(X)
        n_features = X.shape[1]

        # Check from/to causal order
        if from_lag == 0:
            from_order = self._causal_order.index(from_index)
            to_order = self._causal_order.index(to_index)
            if from_order > to_order:
                warnings.warn(f'The estimated causal effect may be incorrect because '
                              f'the causal order of the destination variable (to_index={to_index}) '
                              f'is earlier than the source variable (from_index={from_index}).')

        # X + lagged X
        X_joined = np.zeros((X.shape[0], X.shape[1] * (1 + self._lags + from_lag)))
        for p in range(1 + self._lags + from_lag):
            pos = n_features * p
            X_joined[:, pos:pos +
                            n_features] = np.roll(X[:, 0:n_features], p, axis=0)

        # from_index + parents indices
        am = np.concatenate([*self._adjacency_matrices], axis=1)
        parents = np.where(np.abs(am[from_index]) > 0)[0]
        from_index = from_index if from_lag == 0 else from_index + \
                                                      (n_features * from_lag)
        parents = parents if from_lag == 0 else parents + (n_features * from_lag)
        predictors = [from_index]
        predictors.extend(parents)

        # estimate total effect
        coefs = predict_adaptive_lasso(X_joined, predictors, to_index)

        return coefs[0]

    def get_error_independence_p_values(self):
        """Calculate the p-value matrix of independence between error variables.

        Returns
        -------
        independence_p_values : array-like, shape (n_features, n_features)
            p-value matrix of independence between error variables.
        """
        nn = self.residuals_
        B0 = self._adjacency_matrices[0]
        E = np.dot(np.eye(B0.shape[0]) - B0, nn.T).T
        n_samples = E.shape[0]
        n_features = E.shape[1]

        p_values = np.zeros([n_features, n_features])
        for i, j in itertools.combinations(range(n_features), 2):
            _, p_value = hsic_test_gamma(np.reshape(E[:, i], [n_samples, 1]),
                                         np.reshape(E[:, j], [n_samples, 1]))
            p_values[i, j] = p_value
            p_values[j, i] = p_value

        return p_values

    def _estimate_var_coefs(self, X):
        """Estimate coefficients of VAR"""
        #@todo: VAR.fit() is not searching lags correctly
        if self._criterion not in ['aic', 'fpe', 'hqic', 'bic']:
            var = VAR(X)
            result = var.fit(maxlags=self._lags, trend='n')
        else:
            min_value = float('Inf')
            result = None

            for lag in range(1, self._lags + 1):
                var = VAR(X)
                fitted = var.fit(maxlags=lag, ic=None, trend='n')

                value = getattr(fitted, self._criterion)
                if value < min_value:
                    min_value = value
                    result = fitted

        return result.coefs, result.k_ar, result.resid

    def _calc_residuals(self, X, M_taus, lags):
        """Calculate residuals"""
        X = X.T
        n_features = X.shape[0]
        n_samples = X.shape[1]

        residuals = np.zeros((n_features, n_samples))
        for t in range(n_samples):
            if t - lags < 0:
                continue

            estimated = np.zeros((X.shape[0], 1))
            for tau in range(1, lags + 1):
                estimated += np.dot(M_taus[tau - 1],
                                    X[:, t - tau].reshape((-1, 1)))

            residuals[:, t] = X[:, t] - estimated.reshape((-1,))

        residuals = residuals[:, lags:].T

        return residuals

    def _calc_b(self, X, B0, M_taus):
        """Calculate B"""
        n_features = X.shape[1]

        B_taus = np.array([B0])

        for M in M_taus:
            B_t = np.dot((np.eye(n_features) - B0), M)
            B_taus = np.append(B_taus, [B_t], axis=0)

        return B_taus

    def _pruning(self, X, B_taus, causal_order):
        """Prune edges"""
        n_features = X.shape[1]

        stacked = [np.flip(X, axis=0)]
        for i in range(self._lags):
            stacked.append(np.roll(stacked[-1], -1, axis=0))
        blocks = np.array(list(zip(*stacked)))[:-self._lags]

        for i in range(n_features):
            causal_order_no = causal_order.index(i)
            ancestor_indexes = causal_order[:causal_order_no]

            obj = np.zeros((len(blocks)))
            exp = np.zeros(
                (len(blocks), causal_order_no + n_features * self._lags))
            for j, block in enumerate(blocks):
                obj[j] = block[0][i]
                exp[j:] = np.concatenate(
                    [block[0][ancestor_indexes].flatten(), block[1:][:].flatten()], axis=0)

            # adaptive lasso
            predictors = [i for i in range(exp.shape[1])]
            target = len(predictors)
            X_con = np.concatenate([exp, obj.reshape(-1, 1)], axis=1)
            coef = predict_adaptive_lasso(X_con, predictors, target)

            B_taus[0][i, ancestor_indexes] = coef[:causal_order_no]
            for j in range(len(B_taus[1:])):
                B_taus[j + 1][i, :] = coef[causal_order_no + n_features *
                                           j: causal_order_no + n_features * j + n_features]

        return B_taus

    @property
    def causal_order_(self):
        """Estimated causal ordering.

        Returns
        -------
        causal_order_ : array-like, shape (n_features)
            The causal order of fitted model, where
            n_features is the number of features.
        """
        return self._causal_order

    @property
    def adjacency_matrices_(self):
        """Estimated adjacency matrix.

        Returns
        -------
        adjacency_matrices_ : array-like, shape (lags, n_features, n_features)
            The adjacency matrix of fitted model, where
            n_features is the number of features.
        """
        return self._adjacency_matrices

    @property
    def residuals_(self):
        """Residuals of regression.

        Returns
        -------
        residuals_ : array-like, shape (n_samples)
            Residuals of regression, where n_samples is the number of samples.
        """
        return self._residuals



import numbers
import numpy as np

from sklearn.linear_model import LassoLarsIC

__all__ = ['get_kernel_width', 'get_gram_matrix', 'hsic_teststat', \
           'hsic_test_gamma', 'predict_adaptive_lasso', 'find_all_paths']

import numpy as np
from math import ceil, exp, gamma

def gamma_pdf(x, a, b):
    if x < 0:
        return 0
    else:
        return (x**(a-1) * exp(-x/b)) / (b**a * gamma(a))

def gamma_cdf(x, a, b):
    if x < 0:
        return 0
    else:
        n = int(ceil(x/b))
        h = x / (b * n)
        sum = 0
        for i in range(n):
            x_i = i * h
            sum += 0.5 * (gamma_pdf(x_i, a, b) + gamma_pdf(x_i + h, a, b)) * h
        return sum



def linear_regression(X, y=None, fit_intercept=True):
    # Add intercept term if required
    if fit_intercept:
        X = np.hstack([np.ones((X.shape[0], 1)), X])

    # Compute SVD of X
    U, S, VT = np.linalg.svd(X, full_matrices=False)

    # Compute pseudoinverse of X
    S_inv = np.diag(1 / S)
    X_pseudo_inv = VT.T @ S_inv @ U.T

    # Compute coefficients
    coef = X_pseudo_inv @ y

    return coef


def scale(X):
    """Centers and scales the input features."""
    mean = X.mean(axis=0)
    std = X.std(axis=0)
    X_centered = X - mean
    X_scaled = X_centered / std
    return X_scaled


def resample(X, y, n_samples=None, replace=True, stratify=None, random_state=None):
    """Resample arrays or sparse matrices.

    Parameters:
    ----------
    X : {ndarray, sparse matrix}
        The data to resample. Must be 2D, with the number of samples
        in the first dimension and the number of features in the second dimension.
    y : {ndarray, int, None}
        The labels, if any. Must be 1D, with the number of labels
        equal to the number of samples in X. If None, the labels
        are not resampled.
    n_samples : int, default=None
        The number of samples to resample. If None, the length of X is
        used. If replace is False, the number of samples must be less
        than or equal to the length of X.
    replace : bool, default=True
        Whether to resample with replacement. If True, samples are drawn
        with replacement, which means that the same sample can be drawn
        multiple times. If False, samples are drawn without replacement,
        which means that each sample can only be drawn once.
    stratify : array-like, default=None
        If not None, data is split in a stratified fashion, using this
        as the class labels. This argument is ignored if y is None.
    random_state : int, RandomState instance or None, default=None
        Determines random number generation for shuffling the data.
        Pass an int for reproducible results across multiple function calls.
        See Glossary.

    Returns:
    ----------
    X_resampled : array or sparse matrix
        The resampled data.
    y_resampled : array or None
        The resampled labels, if any.
    """

    if n_samples is None:
        n_samples = X.shape[0]

    if replace and n_samples > X.shape[0]:
        raise ValueError("Number of samples must be less than or equal to the length of X if replace is True.")

    indices = np.random.choice(X.shape[0], size=n_samples, replace=replace, random_state=random_state)

    X_resampled = X[indices]

    if stratify is not None:
        strata, counts = np.unique(stratify, return_counts=True)
        n_samples_per_stratum = np.floor(n_samples * counts / np.sum(counts))

        indices_per_stratum = []
        for stratum, n_samples in zip(strata, n_samples_per_stratum):
            indices_stratum = np.random.choice(np.where(stratify == stratum)[0], size=int(n_samples), replace=False, random_state=random_state)
            indices_per_stratum.append(indices_stratum)

        indices = np.concatenate(indices_per_stratum)

    if y is not None:
        y_resampled = y[indices]

    return X_resampled, y_resampled


def check_array(X, accept_sparse=False, copy=False, ensure_2d=True, allow_nd=False, dtype=None, order=None, force_all_finite=False):
    """Check the array and cast it if necessary.

    Parameters
    ----------
    X : array-like
        The array to check.

    accept_sparse : bool, default=False
        If True, allows sparse matrices.

    copy : bool, default=False
        If True, the data will be copied, otherwise it may be modified
        in-place.

    ensure_2d : bool, default=True
        If True, ensures 2-dimensionality of the input. Otherwise,
        unchanged.

    allow_nd : bool, default=False
        If True, accepts N-dimensional arrays. Otherwise, checks the
        dimensionality of the input.

    dtype : dtype, default=None
        Data type of the result. If None, the data type of the
        input is preserved.

    order : {'C', 'F', 'A'}, default=None
        Order of the result. If None, the order of the input is preserved.

    force_all_finite : bool, default=False
        If True, forces all values of the input to be finite.
        Otherwise, values of the input may be NaN or inf.

    Returns
    -------
    X_new : array
        A checked and cast copy of the array X.

    """

    if hasattr(X, "toarray"):
        # if a sparse matrix, convert it to a dense array
        X = X.toarray(copy=copy)
    else:
        # ensure X is a numpy array
        X = np.asarray(X, order=order)

    if ensure_2d and X.ndim < 2:
        X = X.reshape(1, X.shape[0])

    if force_all_finite:
        X_finite = np.isfinite(X)
        if not X_finite.all():
            raise ValueError("Some of the values in X are not finite.")

    if dtype is not None:
        X = X.astype(dtype, copy=False, order=order)

    return X



def predict_adaptive_lasso(X, predictors, target, gamma=1.0):
    """Predict with Adaptive Lasso.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
        Training data, where n_samples is the number of samples
        and n_features is the number of features.
    predictors : array-like, shape (n_predictors)
        Indices of predictor variable.
    target : int
        Index of target variable.

    Returns
    -------
    coef : array-like, shape (n_features)
        Coefficients of predictor variable.
    """
    # X = X.astype(np.float64)
    coef = linear_regression(X[:, predictors], X[:, target])[1:]
    weight = np.power(np.abs(coef), gamma)

    reg = LassoLarsIC(criterion='bic', max_iter=100)
    reg.fit(X[:, predictors] * weight, X[:, target])

    return reg.coef_ * weight


def find_all_paths(dag, from_index, to_index, min_causal_effect=0.0):
    """Find all paths from point to point in DAG.

    Parameters
    ----------
    dag : array-like, shape (n_features, n_features)
        The adjacency matrix to fine all paths, where n_features is the number of features.
    from_index : int
        Index of the variable at the start of the path.
    to_index : int
        Index of the variable at the end of the path.
    min_causal_effect : float, optional (default=0.0)
        Threshold for detecting causal direction.
        Causal directions with absolute values of causal effects less than ``min_causal_effect`` are excluded.

    Returns
    -------
    paths : array-like, shape (n_paths)
        List of found path, where n_paths is the number of paths.
    effects : array-like, shape (n_paths)
        List of causal effect, where n_paths is the number of paths.
    """
    # Extract all edges
    edges = np.array(np.where(np.abs(np.nan_to_num(dag)) > min_causal_effect)).T

    # Aggregate edges by start point
    to_indices = []
    for i in range(dag.shape[0]):
        adj_list = edges[edges[:, 1] == i][:, 0].tolist()
        if len(adj_list) != 0:
            to_indices.append(adj_list)
        else:
            to_indices.append([])

    # DFS
    paths = []
    stack = [from_index]
    stack_to_indice = [to_indices[from_index]]
    while stack:
        if len(stack) > dag.shape[0]:
            raise ValueError(
                "Unable to find the path because a cyclic graph has been specified.")

        cur_index = stack[-1]
        to_indice = stack_to_indice[-1]

        if cur_index == to_index:
            paths.append(stack.copy())
            stack.pop()
            stack_to_indice.pop()
        else:
            if len(to_indice) > 0:
                next_index = to_indice.pop(0)
                stack.append(next_index)
                stack_to_indice.append(to_indices[next_index].copy())
            else:
                stack.pop()
                stack_to_indice.pop()

    # Calculate the causal effect for each path
    effects = []
    for p in paths:
        coefs = [dag[p[i + 1], p[i]] for i in range(len(p) - 1)]
        effects.append(np.cumprod(coefs)[-1])

    return paths, effects


class BootstrapMixin():
    """Mixin class for all LiNGAM algorithms that implement the method of bootstrapping."""

    def bootstrap(self, X, n_sampling):
        """Evaluate the statistical reliability of DAG based on the bootstrapping.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where ``n_samples`` is the number of samples
            and ``n_features`` is the number of features.
        n_sampling : int
            Number of bootstrapping samples.

        Returns
        -------
        result : BootstrapResult
            Returns the result of bootstrapping.
        """
        # Check parameters
        X = check_array(X)

        if isinstance(n_sampling, (numbers.Integral, np.integer)):
            if not 0 < n_sampling:
                raise ValueError(
                    'n_sampling must be an integer greater than 0.')
        else:
            raise ValueError('n_sampling must be an integer greater than 0.')

        # Bootstrapping
        adjacency_matrices = np.zeros([n_sampling, X.shape[1], X.shape[1]])
        total_effects = np.zeros([n_sampling, X.shape[1], X.shape[1]])
        for i in range(n_sampling):
            self.fit(resample(X))
            adjacency_matrices[i] = self._adjacency_matrix

            # Calculate total effects
            for c, from_ in enumerate(self._causal_order):
                for to in self._causal_order[c + 1:]:
                    total_effects[i, to, from_] = self.estimate_total_effect(
                        X, from_, to)

        return BootstrapResult(adjacency_matrices, total_effects)


class BootstrapResult(object):
    """The result of bootstrapping."""

    def __init__(self, adjacency_matrices, total_effects):
        """Construct a BootstrapResult.

        Parameters
        ----------
        adjacency_matrices : array-like, shape (n_sampling)
            The adjacency matrix list by bootstrapping.
        total_effects : array-like, shape (n_sampling)
            The total effects list by bootstrapping.
        """
        self._adjacency_matrices = adjacency_matrices
        self._total_effects = total_effects

    @property
    def adjacency_matrices_(self):
        """The adjacency matrix list by bootstrapping.

        Returns
        -------
        adjacency_matrices_ : array-like, shape (n_sampling)
            The adjacency matrix list, where ``n_sampling`` is
            the number of bootstrap sampling.
        """
        return self._adjacency_matrices

    @property
    def total_effects_(self):
        """The total effect list by bootstrapping.

        Returns
        -------
        total_effects_ : array-like, shape (n_sampling)
            The total effect list, where ``n_sampling`` is
            the number of bootstrap sampling.
        """
        return self._total_effects

    def get_causal_direction_counts(self, n_directions=None, min_causal_effect=None, split_by_causal_effect_sign=False):
        """Get causal direction count as a result of bootstrapping.

        Parameters
        ----------
        n_directions : int, optional (default=None)
            If int, then The top ``n_directions`` items are included in the result
        min_causal_effect : float, optional (default=None)
            Threshold for detecting causal direction.
            If float, then causal directions with absolute values of causal effects less than ``min_causal_effect`` are excluded.
        split_by_causal_effect_sign : boolean, optional (default=False)
            If True, then causal directions are split depending on the sign of the causal effect.

        Returns
        -------
        causal_direction_counts : dict
            List of causal directions sorted by count in descending order.
            The dictionary has the following format::

            {'from': [n_directions], 'to': [n_directions], 'count': [n_directions]}

            where ``n_directions`` is the number of causal directions.
        """
        # Check parameters
        if isinstance(n_directions, (numbers.Integral, np.integer)):
            if not 0 < n_directions:
                raise ValueError(
                    'n_directions must be an integer greater than 0')
        elif n_directions is None:
            pass
        else:
            raise ValueError('n_directions must be an integer greater than 0')

        if min_causal_effect is None:
            min_causal_effect = 0.0
        else:
            if not 0.0 < min_causal_effect:
                raise ValueError(
                    'min_causal_effect must be an value greater than 0.')

        # Count causal directions
        directions = []
        for am in np.nan_to_num(self._adjacency_matrices):
            direction = np.array(np.where(np.abs(am) > min_causal_effect))
            if split_by_causal_effect_sign:
                signs = np.array([np.sign(am[i][j])
                                  for i, j in direction.T]).astype('int64').T
                direction = np.vstack([direction, signs])
            directions.append(direction.T)
        directions = np.concatenate(directions)

        if len(directions) == 0:
            cdc = {'from': [], 'to': [], 'count': []}
            if split_by_causal_effect_sign:
                cdc['sign'] = []
            return cdc

        directions, counts = np.unique(directions, axis=0, return_counts=True)
        sort_order = np.argsort(-counts)
        sort_order = sort_order[:n_directions] if n_directions is not None else sort_order
        counts = counts[sort_order]
        directions = directions[sort_order]

        cdc = {
            'from': directions[:, 1].tolist(),
            'to': directions[:, 0].tolist(),
            'count': counts.tolist()
        }
        if split_by_causal_effect_sign:
            cdc['sign'] = directions[:, 2].tolist()

        return cdc

    def get_directed_acyclic_graph_counts(self, n_dags=None, min_causal_effect=None, split_by_causal_effect_sign=False):
        """Get DAGs count as a result of bootstrapping.

        Parameters
        ----------
        n_dags : int, optional (default=None)
            If int, then The top ``n_dags`` items are included in the result
        min_causal_effect : float, optional (default=None)
            Threshold for detecting causal direction.
            If float, then causal directions with absolute values of causal effects less than ``min_causal_effect`` are excluded.
        split_by_causal_effect_sign : boolean, optional (default=False)
            If True, then causal directions are split depending on the sign of the causal effect.

        Returns
        -------
        directed_acyclic_graph_counts : dict
            List of directed acyclic graphs sorted by count in descending order.
            The dictionary has the following format::

            {'dag': [n_dags], 'count': [n_dags]}.

            where ``n_dags`` is the number of directed acyclic graphs.
        """
        # Check parameters
        if isinstance(n_dags, (numbers.Integral, np.integer)):
            if not 0 < n_dags:
                raise ValueError('n_dags must be an integer greater than 0')
        elif n_dags is None:
            pass
        else:
            raise ValueError('n_dags must be an integer greater than 0')

        if min_causal_effect is None:
            min_causal_effect = 0.0
        else:
            if not 0.0 < min_causal_effect:
                raise ValueError(
                    'min_causal_effect must be an value greater than 0.')

        # Count directed acyclic graphs
        dags = []
        for am in np.nan_to_num(self._adjacency_matrices):
            dag = np.abs(am) > min_causal_effect
            if split_by_causal_effect_sign:
                direction = np.array(np.where(dag))
                signs = np.zeros_like(dag).astype('int64')
                for i, j in direction.T:
                    signs[i][j] = np.sign(am[i][j]).astype('int64')
                dag = signs
            dags.append(dag)

        dags, counts = np.unique(dags, axis=0, return_counts=True)
        sort_order = np.argsort(-counts)
        sort_order = sort_order[:n_dags] if n_dags is not None else sort_order
        counts = counts[sort_order]
        dags = dags[sort_order]

        if split_by_causal_effect_sign:
            dags = [{
                'from': np.where(dag)[1].tolist(),
                'to': np.where(dag)[0].tolist(),
                'sign': [dag[i][j] for i, j in np.array(np.where(dag)).T]} for dag in dags]
        else:
            dags = [{
                'from': np.where(dag)[1].tolist(),
                'to': np.where(dag)[0].tolist()} for dag in dags]

        return {
            'dag': dags,
            'count': counts.tolist()
        }

    def get_probabilities(self, min_causal_effect=None):
        """Get bootstrap probability.

        Parameters
        ----------
        min_causal_effect : float, optional (default=None)
            Threshold for detecting causal direction.
            If float, then causal directions with absolute values of causal effects less than ``min_causal_effect`` are excluded.

        Returns
        -------
        probabilities : array-like
            List of bootstrap probability matrix.
        """
        # check parameters
        if min_causal_effect is None:
            min_causal_effect = 0.0
        else:
            if not 0.0 < min_causal_effect:
                raise ValueError(
                    'min_causal_effect must be an value greater than 0.')

        adjacency_matrices = np.nan_to_num(self._adjacency_matrices)
        shape = adjacency_matrices[0].shape
        bp = np.zeros(shape)
        for B in adjacency_matrices:
            bp += np.where(np.abs(B) > min_causal_effect, 1, 0)
        bp = bp / len(adjacency_matrices)

        if int(shape[1] / shape[0]) == 1:
            return bp
        else:
            return np.hsplit(bp, int(shape[1] / shape[0]))

    def get_total_causal_effects(self, min_causal_effect=None):
        """Get total effects list.

        Parameters
        ----------
        min_causal_effect : float, optional (default=None)
            Threshold for detecting causal direction.
            If float, then causal directions with absolute values of causal effects less than ``min_causal_effect`` are excluded.

        Returns
        -------
        total_causal_effects : dict
            List of bootstrap total causal effect sorted by probability in descending order.
            The dictionary has the following format::

            {'from': [n_directions], 'to': [n_directions], 'effect': [n_directions], 'probability': [n_directions]}

            where ``n_directions`` is the number of causal directions.
        """
        # Check parameters
        if min_causal_effect is None:
            min_causal_effect = 0.0
        else:
            if not 0.0 < min_causal_effect:
                raise ValueError(
                    'min_causal_effect must be an value greater than 0.')

        # Calculate probability
        probs = np.sum(np.where(np.abs(self._total_effects) >
                                min_causal_effect, 1, 0), axis=0, keepdims=True)[0]
        probs = probs / len(self._total_effects)

        # Causal directions
        dirs = np.array(np.where(np.abs(probs) > 0))
        probs = probs[dirs[0], dirs[1]]

        # Calculate median effect without zero
        effects = np.zeros(dirs.shape[1])
        for i, (to, from_) in enumerate(dirs.T):
            idx = np.where(np.abs(self._total_effects[:, to, from_]) > 0)
            effects[i] = np.median(self._total_effects[:, to, from_][idx])

        # Sort by probability
        order = np.argsort(-probs)
        dirs = dirs.T[order]
        effects = effects[order]
        probs = probs[order]

        ce = {
            'from': dirs[:, 1].tolist(),
            'to': dirs[:, 0].tolist(),
            'effect': effects.tolist(),
            'probability': probs.tolist()
        }

        return ce

    def get_paths(self, from_index, to_index, min_causal_effect=0.0):
        """Get all paths from the start variable to the end variable and their bootstrap probabilities.

        Parameters
        ----------
        from_index : int
            Index of the variable at the start of the path.
        to_index : int
            Index of the variable at the end of the path.
        min_causal_effect : float, optional (default=0.0)
            Threshold for detecting causal direction.
            Causal directions with absolute values of causal effects less than ``min_causal_effect`` are excluded.

        Returns
        -------
        paths : dict
            List of path and bootstrap probability.
            The dictionary has the following format::

            {'path': [n_paths], 'effect': [n_paths], 'probability': [n_paths]}

            where ``n_paths`` is the number of paths.
        """
        # Find all paths from from_index to to_index
        paths_list = []
        effects_list = []
        for am in self._adjacency_matrices:
            paths, effects = find_all_paths(am, from_index, to_index)
            # Convert path to string to make them easier to handle.
            paths_list.extend(['_'.join(map(str, p)) for p in paths])
            effects_list.extend(effects)

        paths_list = np.array(paths_list)
        effects_list = np.array(effects_list)

        # Count paths
        paths_str, counts = np.unique(paths_list, axis=0, return_counts=True)

        # Sort by count
        order = np.argsort(-counts)
        probs = counts[order] / len(self._adjacency_matrices)
        paths_str = paths_str[order]

        # Calculate median of causal effect for each path
        effects = [np.median(effects_list[np.where(paths_list == p)])
                   for p in paths_str]

        result = {
            'path': [[int(i) for i in p.split('_')] for p in paths_str],
            'effect': effects,
            'probability': probs.tolist(),
        }
        return result


def get_kernel_width(X):
    """Calculate the bandwidth to median distance between points.
    Use at most 100 points (since median is only a heuristic,
    and 100 points is sufficient for a robust estimate).

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
        Training data, where ``n_samples`` is the number of samples
        and ``n_features`` is the number of features.

    Returns
    -------
    float
        The bandwidth parameter.
    """
    n_samples = X.shape[0]
    if n_samples > 100:
        X_med = X[:100, :]
        n_samples = 100
    else:
        X_med = X

    G = np.sum(X_med * X_med, 1).reshape(n_samples, 1)
    dists = G + G.T - 2 * np.dot(X_med, X_med.T)
    dists = dists - np.tril(dists)
    dists = dists.reshape(n_samples ** 2, 1)

    return np.sqrt(0.5 * np.median(dists[dists > 0]))

def _rbf_dot(X, Y, width):
    """Compute the inner product of radial basis functions."""
    n_samples_X = X.shape[0]
    n_samples_Y = Y.shape[0]

    G = np.sum(X * X, 1).reshape(n_samples_X, 1)
    H = np.sum(Y * Y, 1).reshape(n_samples_Y, 1)
    Q = np.tile(G, (1, n_samples_Y))
    R = np.tile(H.T, (n_samples_X, 1))
    H = Q + R - 2 * np.dot(X, Y.T)

    return np.exp(-H / 2 / (width ** 2))

def _rbf_dot_XX(X, width):
    """rbf dot, in special case with X dot X"""
    G = np.sum(X * X, axis=1)
    H = G[None, :] + G[:, None] - 2 * np.dot(X, X.T)
    return np.exp(-H / 2 / (width ** 2))

def get_gram_matrix(X, width):
    """Get the centered gram matrices.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
        Training data, where ``n_samples`` is the number of samples
        and ``n_features`` is the number of features.

    width : float
        The bandwidth parameter.

    Returns
    -------
    K, Kc : array
        the centered gram matrices.
    """
    n = X.shape[0]

    K = _rbf_dot_XX(X, width)
    K_colsums = K.sum(axis=0)
    K_rowsums = K.sum(axis=1)
    K_allsum = K_rowsums.sum()
    Kc = K - (K_colsums[None, :] + K_rowsums[:, None]) / n + (K_allsum / n ** 2)
    # equivalent to H @ K @ H, where H = np.eye(n) - 1 / n * np.ones((n, n)).
    return K, Kc


def hsic_teststat(Kc, Lc, n):
    """get the HSIC statistic.

    Parameters
    ----------
    K, Kc : array
        the centered gram matrices.

    n : float
        the number of samples.

    Returns
    -------
    float
        the HSIC statistic.
    """
    # test statistic m*HSICb under H1
    return 1 / n * np.sum(Kc.T * Lc)


def hsic_test_gamma(X, Y, bw_method='mdbs'):
    """get the HSIC statistic.

    Parameters
    ----------
    X, Y : array-like, shape (n_samples, n_features)
        Training data, where ``n_samples`` is the number of samples
        and ``n_features`` is the number of features.

    bw_method : str, optional (default=``mdbs``)
        The method used to calculate the bandwidth of the HSIC.

        * ``mdbs`` : Median distance between samples.
        * ``scott`` : Scott's Rule of Thumb.
        * ``silverman`` : Silverman's Rule of Thumb.

    Returns
    -------
    test_stat : float
        the HSIC statistic.

    p : float
        the HSIC p-value.
    """
    X = X.reshape(-1, 1) if X.ndim == 1 else X
    Y = Y.reshape(-1, 1) if Y.ndim == 1 else Y

    # Get kernel width to median distance between points
    width_x = get_kernel_width(X)
    width_y = get_kernel_width(Y)

    # these are slightly biased estimates of centered gram matrices
    K, Kc = get_gram_matrix(X, width_x)
    L, Lc = get_gram_matrix(Y, width_y)

    # test statistic m*HSICb under H1
    n = X.shape[0]
    test_stat = hsic_teststat(Kc, Lc, n)

    var = (1 / 6 * Kc * Lc) ** 2
    # second subtracted term is bias correction
    var = 1 / n / (n - 1) * (np.sum(var) - np.trace(var))
    # variance under H0
    var = 72 * (n - 4) * (n - 5) / n / (n - 1) / (n - 2) / (n - 3) * var

    K[np.diag_indices(n)] = 0
    L[np.diag_indices(n)] = 0
    mu_X = 1 / n / (n - 1) * K.sum()
    mu_Y = 1 / n / (n - 1) * L.sum()
    # mean under H0
    mean = 1 / n * (1 + mu_X * mu_Y - mu_X - mu_Y)

    alpha = mean ** 2 / var
    # threshold for hsicArr*m
    beta = var * n / mean
    p = 1 - gamma_cdf(test_stat, alpha, beta)

    return test_stat, p



import ctypes
import numpy as np
cuda_lib = ctypes.CDLL('./unit.so') #check that path exists
import time

def test1():
    # Define the wrapper function
    run_compute_mean = cuda_lib.run_compute_mean
    run_compute_mean.argtypes = [np.ctypeslib.ndpointer(dtype=np.float64),
                                np.ctypeslib.ndpointer(dtype=np.float64),
                                ctypes.c_int,
                                ctypes.c_int]


    size = 1000000
    blockSize = 256

    # Generate dummy data
    data = np.random.rand(size).astype(np.float64)
    mean = np.zeros(1, dtype=np.float64)

    # Call the wrapper function
    run_compute_mean(data, mean, size, blockSize)

    # Calculate expected mean using NumPy
    expected_mean = np.mean(data)

    # Check the result
    # print(mean[0], expected_mean);
    assert np.isclose(mean[0], expected_mean), f"Result {mean[0]} does not match expected {expected_mean}"

    print("Test passed")


def test2():
    run_compute_covariance_variance = cuda_lib.run_compute_covariance_variance
    run_compute_covariance_variance.argtypes = [np.ctypeslib.ndpointer(dtype=np.float64),
                                            np.ctypeslib.ndpointer(dtype=np.float64),
                                            np.ctypeslib.ndpointer(dtype=np.float64),
                                            np.ctypeslib.ndpointer(dtype=np.float64),
                                            np.ctypeslib.ndpointer(dtype=np.float64),
                                            np.ctypeslib.ndpointer(dtype=np.float64),
                                            ctypes.c_int,
                                            ctypes.c_int]

    size = 4177
    blockSize = 256

    xi = np.random.rand(size).astype(np.float64)
    xj = np.random.rand(size).astype(np.float64)
    mean_xi = np.array([np.mean(xi)], dtype=np.float64)
    mean_xj = np.array([np.mean(xj)], dtype=np.float64)
    covariance = np.zeros(1, dtype=np.float64)
    variance = np.zeros(1, dtype=np.float64)

    # Call the wrapper function
    run_compute_covariance_variance(xi, xj, mean_xi, mean_xj, covariance, variance, size, blockSize)

    # Calculate expected covariance and variance using NumPy
    expected_covariance = np.sum((xi - mean_xi) * (xj - mean_xj))
    expected_variance = np.sum((xj - mean_xj)**2)

    # print(covariance[0], expected_covariance)
    # print(variance[0], expected_variance)
    assert np.isclose(covariance[0], expected_covariance), "Covariance test failed {}, {}".format(covariance[0], expected_covariance)
    assert np.isclose(variance[0], expected_variance), "Variance test failed"

    print("Test passed")

def test3():
    run_element_wise_division = cuda_lib.run_element_wise_division
    run_element_wise_division.argtypes = [np.ctypeslib.ndpointer(dtype=np.float64),
                                        np.ctypeslib.ndpointer(dtype=np.float64),
                                        np.ctypeslib.ndpointer(dtype=np.float64),
                                        ctypes.c_int,
                                        ctypes.c_int]

    n = 4177
    blockSize = 256

    r = np.random.rand(n).astype(np.float64)
    constant_std = np.array([np.random.rand()], dtype=np.float64)
    result = np.zeros(n, dtype=np.float64)

    # Call the wrapper function
    run_element_wise_division(r, constant_std, result, n, blockSize)

    # Calculate expected result using NumPy
    expected_result = np.divide(r, constant_std) if constant_std != 0 else np.zeros(n, dtype=np.float64)

    assert np.allclose(result, expected_result), "element_wise_division Test failed"

    print("Test passed")

def test4():
    # Define the wrapper function
    run_compute_residual = cuda_lib.run_compute_residual
    run_compute_residual.argtypes = [np.ctypeslib.ndpointer(dtype=np.float64),
                                    np.ctypeslib.ndpointer(dtype=np.float64),
                                    np.ctypeslib.ndpointer(dtype=np.float64),
                                    np.ctypeslib.ndpointer(dtype=np.float64),
                                    ctypes.c_int,
                                    ctypes.c_int]

    size = 4177
    blockSize = 256

    xi = np.random.rand(size).astype(np.float64)
    xj = np.random.rand(size).astype(np.float64)
    scaling_factor = np.array([np.random.rand()], dtype=np.float64)
    residual = np.zeros(size, dtype=np.float64)

    # Call the wrapper function
    run_compute_residual(xi, xj, scaling_factor, residual, size, blockSize)

    # Calculate expected result using NumPy
    expected_residual = xi - scaling_factor * xj

    assert np.allclose(residual, expected_residual), "compute_residual Test failed"

    print("Test passed")

def test5():
    run_compute_std = cuda_lib.run_compute_std
    run_compute_std.argtypes = [np.ctypeslib.ndpointer(dtype=np.float64),
                                np.ctypeslib.ndpointer(dtype=np.float64),
                                np.ctypeslib.ndpointer(dtype=np.float64),
                                ctypes.c_int,
                                ctypes.c_int]

    size = 4177
    blockSize = 256

    A = np.random.rand(size).astype(np.float64)
    mean = np.array([np.mean(A)], dtype=np.float64)
    std = np.array([0.0], dtype=np.float64)

    # Call the wrapper function
    run_compute_std(A, mean, std, size, blockSize)

    # Calculate expected standard deviation using NumPy
    expected_std = np.std(A)
    # print(std[0], expected_std)
    assert np.isclose(std[0], expected_std), "compute_std Test failed"

    print("Test passed")


def test6():
    run_calculate_statistics = cuda_lib.run_calculate_statistics
    run_calculate_statistics.argtypes = [np.ctypeslib.ndpointer(dtype=np.float64),
                                        ctypes.c_int,
                                        ctypes.c_int,
                                        np.ctypeslib.ndpointer(dtype=np.float64),
                                        np.ctypeslib.ndpointer(dtype=np.float64),
                                        ctypes.c_int]

    m, n = 4177, 11
    blockSize = 256

    A = np.random.rand(m, n).astype(np.float64)
    means = np.zeros(n, dtype=np.float64)
    stds = np.zeros(n, dtype=np.float64)

    # Call the wrapper function
    run_calculate_statistics(A, m, n, means, stds, blockSize)

    # Calculate expected results using NumPy
    expected_means = np.mean(A, axis=0)
    expected_stds = np.std(A, axis=0)

    assert np.allclose(means, expected_means), "Means test failed"
    assert np.allclose(stds, expected_stds), "Standard deviations test failed"

    print("Test passed")

def test7():
    run_standardize_column = cuda_lib.run_standardize_column
    run_standardize_column.argtypes = [np.ctypeslib.ndpointer(dtype=np.float64),
                                    ctypes.c_int,
                                    ctypes.c_int,
                                    np.ctypeslib.ndpointer(dtype=np.float64),
                                    np.ctypeslib.ndpointer(dtype=np.float64),
                                    ctypes.c_int]

    m, n = 417700, 10
    unitthreads = 16

    A = np.random.rand(m, n).astype(np.float64)
    means = np.mean(A, axis=0).astype(np.float64)
    stds = np.std(A, axis=0).astype(np.float64)

    expected_A = np.zeros((m, n))
    import time
    start_time = time.time()
    for k in range(n):
        expected_A[:, k] = (A[:, k] - np.mean(A[:, k])) / np.std(A[:, k]);
    time_for_loop = time.time() - start_time
    start_time = time.time()
    run_standardize_column(A, m, n, means, stds, unitthreads)
    time_run_standardize_column = time.time() - start_time

    assert np.allclose(A, expected_A), "Test failed"

    # print("Test passed", time_for_loop, time_run_standardize_column)
    print("Test passed")


def test8():
    run_compute_log_cosh = cuda_lib.run_compute_log_cosh
    run_compute_log_cosh.argtypes = [np.ctypeslib.ndpointer(dtype=np.float64),
                                    np.ctypeslib.ndpointer(dtype=np.float64),
                                    ctypes.c_int,
                                    ctypes.c_int]
    size = 4177
    blockSize = 256

    u = np.random.rand(size).astype(np.float64) * 10
    log_cosh_sum = np.array([0.0], dtype=np.float64)
    run_compute_log_cosh(u, log_cosh_sum, size, blockSize)
    expected_log_cosh_sum = np.sum(np.log(np.cosh(u)))

    # print(log_cosh_sum[0], expected_log_cosh_sum);
    assert np.isclose(log_cosh_sum[0], expected_log_cosh_sum), "Test failed"

    print("Test passed")

def test9():
    run_compute_u_exp = cuda_lib.run_compute_u_exp
    run_compute_u_exp.argtypes = [np.ctypeslib.ndpointer(dtype=np.float64),
                                np.ctypeslib.ndpointer(dtype=np.float64),
                                ctypes.c_int,
                                ctypes.c_int]

    size = 417700
    blockSize = 256

    u = np.random.rand(size).astype(np.float64) * 10  # Random values scaled
    u_exp_sum = np.array([0.0], dtype=np.float64)
    run_compute_u_exp(u, u_exp_sum, size, blockSize)
    expected_u_exp_sum = np.sum(u * np.exp(-0.5 * u * u))

    # print(u_exp_sum[0], expected_u_exp_sum);
    assert np.isclose(u_exp_sum[0], expected_u_exp_sum), "Test failed"

    print("Test passed")


def test10():
    run_process_column = cuda_lib.run_process_column
    run_process_column.argtypes = [ctypes.POINTER(ctypes.c_double),
                                ctypes.POINTER(ctypes.c_double),
                                ctypes.c_int,
                                ctypes.c_int,
                                ctypes.c_int,
                                ctypes.c_int]

    m, n = 4177, 11
    col_idx = 2
    threads = 256

    d_X = np.random.rand(m, n).astype(np.float64)
    column =  np.zeros(m, dtype=np.float64)

    expected_column = d_X[:, col_idx]

    c_array = d_X.ctypes.data_as(ctypes.POINTER(ctypes.c_double))
    d_column = column.ctypes.data_as(ctypes.POINTER(ctypes.c_double))

    run_process_column(c_array, d_column, m, n, col_idx, threads)
    column = np.frombuffer(ctypes.cast(d_column, ctypes.POINTER(ctypes.c_double * m)).contents, dtype=np.float64)
    assert np.allclose(column, expected_column), "Test failed"

    print("Test passed")


def test11():
    end_end_residual = cuda_lib.end_end_residual
    end_end_residual.argtypes = [np.ctypeslib.ndpointer(ctypes.c_double),
                                ctypes.c_int,
                                ctypes.c_int,
                                ctypes.c_int,
                                np.ctypeslib.ndpointer(ctypes.c_int), ctypes.c_int]

    def _residual(xi, xj):
            return xi - (np.cov(xi, xj, ddof=1)[0, 1] / np.var(xj, ddof=1)) * xj

    m, n = 1000, 110
    col_idx = 2
    threads = 256

    d_X = np.random.rand(m, n).astype(np.float64)
    U = np.array(range(n)).astype(np.int32)

    expected_arr = d_X.copy()

    for i in U:
        if i != col_idx:
            expected_arr[:, i] = _residual(expected_arr[:, i], expected_arr[:, col_idx])

    end_end_residual(d_X, m, n, col_idx, U, n)
    d_X = d_X.reshape(m, n)

    # print(d_X, expected_arr);
    assert np.allclose(d_X, expected_arr,), "Test failed"
    print("Test passed")


def test12():
    cuda_causal_order = cuda_lib.causal_order
    cuda_causal_order.argtypes = [ctypes.POINTER(ctypes.c_double), ctypes.c_int, ctypes.c_int]

    X = np.random.rand(1000, 10);
    U = np.array(range(10));

    cols = len(U)
    rows = len(X)

    arr = np.ascontiguousarray(X[:, np.array(U)])
    arr = arr.ravel(order='C')

    U = U.astype(np.float64)
    from lingam_cuda import causal_order

    res = causal_order(X, rows, cols)

    c_array = arr.ctypes.data_as(ctypes.POINTER(ctypes.c_double))

    output_buffer = (ctypes.c_double * cols)()
    cuda_causal_order(c_array, rows, cols, output_buffer)
    M_list = np.frombuffer(output_buffer, dtype=np.float64)
    print(res)
    assert np.allclose(res, M_list), "Test failed"
    print("Test passed")

if __name__ == "__main__":
    test1()
    test2()
    test3()
    test4()
    test5()
    test6()
    test7()
    test8()
    test9()
    test10()
    test11()
    test12()



from setuptools import setup, find_packages
from setuptools_cuda_cpp import CUDAExtension, BuildExtension
import os
import numpy
from pathlib import Path

cuda_home = os.environ.get('CUDA_HOME', '/usr/local/cuda-12.2')
cuda_include_dir = os.path.join(cuda_home, 'include')
cuda_lib_dir = os.path.join(cuda_home, 'lib64')
gpu_arch = os.environ.get('GPU_ARCH', 'sm_86')

def append_nvcc_threads(nvcc_extra_args):
    return nvcc_extra_args + ['--threads', '4']

cuda_ext_path = Path('src/culingam')
cuda_ext = CUDAExtension(
    name='lingam_cuda',
    include_dirs=[cuda_ext_path / 'include', cuda_include_dir, numpy.get_include()],
    sources=[
        str(cuda_ext_path / 'basic.cu'),
        str(cuda_ext_path / 'basic_wrapper.cpp'),
    ],
    libraries=['cudart', 'cudadevrt', 'nvToolsExt'],
    extra_compile_args={
        'cxx': ['-g', '-std=c++17'],
        'nvcc': append_nvcc_threads([
            '-O3',
            '-std=c++17',
            f'-arch={gpu_arch}'
        ])
    },
)

setup(
    name='culingam',
    version='0.0.3',
    author='Victor Akinwande',
    description='CULiNGAM accelerates LiNGAM analysis on GPUs.',
    long_description=open('readme.md').read(),
    long_description_content_type='text/markdown',
    packages=find_packages(where='src'),
    package_dir={'': 'src'},
    ext_modules=[cuda_ext],
    cmdclass={'build_ext': BuildExtension}
)
